SystemOrganization addCategory: #Base32!
SystemOrganization addCategory: #'Base32-Tests'!

!String methodsFor: '*base32' stamp: 'UdoSchneider 10/10/2014 21:26'!
base32Decoded
	"Decode the receiver from base 32"

	^ Base32MimeConverter mimeDecode: self as: self class! !

!String methodsFor: '*base32' stamp: 'UdoSchneider 10/10/2014 21:26'!
base32Encoded
	"Encode the receiver as base32"

	^ (Base32MimeConverter mimeEncode: self readStream) contents! !

!ByteArray methodsFor: '*base32' stamp: 'UdoSchneider 10/10/2014 21:26'!
base32Encoded
	"Encode the receiver as base 32"

	^ (Base32MimeConverter mimeEncode: self readStream) contents! !

TestCase subclass: #Base32MimeConverterTest
	instanceVariableNames: 'message'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Base32-Tests'!

!Base32MimeConverterTest methodsFor: 'tests' stamp: 'UdoSchneider 10/10/2014 23:01'!
assertString: from equalsBase32: to
	| encoded |
	encoded := Base32MimeConverter mimeEncode: from readStream.
	self assert: encoded contents equals: to.
	self assert: (Base32MimeConverter mimeDecodeToChars: encoded) contents equals: from! !

!Base32MimeConverterTest methodsFor: 'setup' stamp: 'UdoSchneider 10/10/2014 21:18'!
setUp
	message := 'Hi There!!' readStream.! !

!Base32MimeConverterTest methodsFor: 'tests' stamp: 'UdoSchneider 10/10/2014 23:01'!
testBase32Encoded
	"self run: #testBase32Encoded"
	
	| encoded |
	encoded := (Base32MimeConverter mimeEncode: message) contents.
	self assert: encoded equals: 'Hi There!!' base32Encoded.
	! !

!Base32MimeConverterTest methodsFor: 'tests' stamp: 'UdoSchneider 10/10/2014 23:05'!
testMimeEncodeDecode
	"self run: #testMimeEncodeDecode"
	
	| encoded |
	encoded := Base32MimeConverter mimeEncode: message.
	self assert: encoded contents equals: 'JBUSAVDIMVZGKII='.
      self assert: (Base32MimeConverter mimeDecodeToChars: encoded) contents equals: message contents.

	"Encoding should proceed from the current stream position."
	message reset.
	message skip: 2.
	encoded := Base32MimeConverter mimeEncode: message.
	self assert: encoded contents equals: 'EBKGQZLSMUQQ===='! !

!Base32MimeConverterTest methodsFor: 'tests' stamp: 'UdoSchneider 10/10/2014 23:02'!
testOnByteArray
	"self run: #testOnByteArray"
	
	self assert: 'Hi There!!' base32Encoded equals: 'Hi There!!' asByteArray base32Encoded! !

!Base32MimeConverterTest methodsFor: 'tests' stamp: 'UdoSchneider 10/10/2014 21:35'!
testRfc4648Vectors
	self
		assertString: '' equalsBase32: '';
		assertString: 'f' equalsBase32: 'MY======';
		assertString: 'fo' equalsBase32: 'MZXQ====';
		assertString: 'foo' equalsBase32: 'MZXW6===';
		assertString: 'foob' equalsBase32: 'MZXW6YQ=';
		assertString: 'fooba' equalsBase32: 'MZXW6YTB';
		assertString: 'foobar' equalsBase32: 'MZXW6YTBOI======'! !

MimeConverter subclass: #Base32MimeConverter
	instanceVariableNames: 'data multiLine'
	classVariableNames: 'FromCharTable ToCharTable'
	poolDictionaries: ''
	category: 'Base32'!

Base32MimeConverter subclass: #Base32HexMimeConverter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Base32'!

!Base32HexMimeConverter class methodsFor: 'class initialization' stamp: 'UdoSchneider 10/29/2014 21:15'!
initialize
	FromCharTable := Array new: 256.	"nils"
	ToCharTable := Array new: 64.
	($0 asciiValue to: $9 asciiValue)
		doWithIndex: [ :val :ind | 
			FromCharTable at: val + 1 put: ind - 1.
			ToCharTable at: ind put: val asCharacter ].
	($A asciiValue to: $V asciiValue)
		doWithIndex: [ :val :ind | 
			FromCharTable at: val + 1 put: ind + 9.
			ToCharTable at: ind + 10 put: val asCharacter ].
	FromCharTable at: $+ asciiValue + 1 put: 62.
	ToCharTable at: 63 put: $+.
	FromCharTable at: $/ asciiValue + 1 put: 63.
	ToCharTable at: 64 put: $/! !

!Base32MimeConverter class methodsFor: 'as yet unclassified' stamp: 'UdoSchneider 10/10/2014 21:36'!
decodeInteger: mimeString 
	"Decode the MIME string into an integer of any length"
	| bytes sum |
	bytes := (Base32MimeConverter mimeDecodeToBytes: mimeString readStream) contents.
	sum := 0.
	bytes reverseDo: [ :by | sum := sum * 256 + by ].
	^ sum! !

!Base32MimeConverter class methodsFor: 'as yet unclassified' stamp: 'UdoSchneider 10/10/2014 21:18'!
encodeInteger: int
	| strm |
	"Encode an integer of any length and return the MIME string"

	strm := ReadWriteStream on: (ByteArray new: int digitLength).
	1 to: int digitLength do: [:ii | strm nextPut: (int digitAt: ii)].
	strm reset.
	^ ((self mimeEncode: strm) contents) copyUpTo: $=	"remove padding"! !

!Base32MimeConverter class methodsFor: 'class initialization' stamp: 'UdoSchneider 10/10/2014 21:41'!
initialize
	FromCharTable := Array new: 256.	"nils"
	ToCharTable := Array new: 64.
	($A asciiValue to: $Z asciiValue)
		doWithIndex: [ :val :ind | 
			FromCharTable at: val + 1 put: ind - 1.
			ToCharTable at: ind put: val asCharacter ].
	($2 asciiValue to: $7 asciiValue)
		doWithIndex: [ :val :ind | 
			FromCharTable at: val + 1 put: ind + 25.
			ToCharTable at: ind + 26 put: val asCharacter ].
	FromCharTable at: $+ asciiValue + 1 put: 62.
	ToCharTable at: 63 put: $+.
	FromCharTable at: $/ asciiValue + 1 put: 63.
	ToCharTable at: 64 put: $/! !

!Base32MimeConverter class methodsFor: 'as yet unclassified' stamp: 'UdoSchneider 10/10/2014 21:37'!
mimeDecodeToBytes: aStream 
	"Return a RWBinaryOrTextStream of the original ByteArray.  aStream has only 33 innocuous character values.  aStream is not binary.  (See class comment). 8 bytes in aStream goes to 3 bytes in output."

	| me |
	aStream position: 0.
	me := self new mimeStream: aStream.
	me dataStream: (RWBinaryOrTextStream on: (ByteArray new: aStream size * 3 // 2)).
	me mimeDecodeToByteArray.
	me dataStream position: 0.
	^ me dataStream! !

!Base32MimeConverter class methodsFor: 'as yet unclassified' stamp: 'UdoSchneider 10/10/2014 21:38'!
mimeDecodeToChars: aStream 
	"Return a ReadWriteStream of the original String.  aStream has only 33 innocuous character values.  It is not binary.  (See class comment). 8 bytes in aStream goes to 3 bytes in output."

	| me |
	aStream position: 0.
	me := self new mimeStream: aStream.
	me dataStream: (ReadWriteStream on: (String new: aStream size * 3 // 2)).
	me mimeDecode.
	me dataStream position: 0.
	^ me dataStream! !

!Base32MimeConverter class methodsFor: 'convenience' stamp: 'UdoSchneider 10/10/2014 21:18'!
mimeEncode: aStream
	"Return a ReadWriteStream of characters.  The data of aStream is encoded as 65 innocuous characters.  (See class comment). 3 bytes in aStream goes to 4 bytes in output."

	^self mimeEncode: aStream multiLine: true! !

!Base32MimeConverter class methodsFor: 'as yet unclassified' stamp: 'UdoSchneider 10/10/2014 21:38'!
mimeEncode: aStream multiLine: aBool
	"Return a ReadWriteStream of characters.  The data of aStream is encoded as 33 innocuous characters.  (See class comment). 3 bytes in aStream goes to 8 bytes in output."

	| me |
	me := self new dataStream: aStream.
	me multiLine: aBool.
	me mimeStream: (ReadWriteStream on: (String new: aStream size + 20 * 4 // 2)).
	me mimeEncode.
	me mimeStream position: 0.
	^ me mimeStream! !

!Base32MimeConverter methodsFor: 'initialization' stamp: 'UdoSchneider 10/10/2014 21:18'!
initialize 
	super initialize.
		
	multiLine := true.! !

!Base32MimeConverter methodsFor: 'conversion' stamp: 'UdoSchneider 10/10/2014 22:01'!
mimeDecode
	"Convert a stream in base 32 with only A-Z,2-7,+,/ to a full byte stream of characters.  Reutrn a whole stream for the user to read."

	| nibA nibB nibC nibD nibE nibF nibG nibH |
	[ mimeStream atEnd ]
		whileFalse: [ 
			(nibA := self nextValue) ifNil: [ ^ dataStream ].
			(nibB := self nextValue) ifNil: [ ^ dataStream ].
			dataStream nextPut: ((nibA bitShift: 3) + (nibB bitShift: -2)) asCharacter.
			nibB := nibB bitAnd: 16r3.
			(nibC := self nextValue) ifNil: [ ^ dataStream ].
			(nibD := self nextValue) ifNil: [ ^ dataStream ].
			dataStream nextPut: ((nibB bitShift: 6) + (nibC bitShift: 1) + (nibD bitShift: -4)) asCharacter.
			nibD := nibD bitAnd: 16rF.
			(nibE := self nextValue) ifNil: [ ^ dataStream ].
			dataStream nextPut: ((nibD bitShift: 4) + (nibE bitShift: -1)) asCharacter.
			nibE := nibE bitAnd: 16r1.
			(nibF := self nextValue) ifNil: [ ^ dataStream ].
			(nibG := self nextValue) ifNil: [ ^ dataStream ].
			dataStream nextPut: ((nibE bitShift: 7) + (nibF bitShift: 2) + (nibG bitShift: -3)) asCharacter.
			nibG := nibG bitAnd: 16r7.
			(nibH := self nextValue) ifNil: [ ^ dataStream ].
			dataStream nextPut: ((nibG bitShift: 5) + nibH) asCharacter ].
	^ dataStream! !

!Base32MimeConverter methodsFor: 'conversion' stamp: 'UdoSchneider 10/10/2014 21:18'!
mimeDecodeToByteArray
	"Convert a stream in base 64 with only a-z,A-Z,0-9,+,/ to a full ByteArray of 0-255 values.  Reutrn a whole stream for the user to read."

	| nibA nibB nibC nibD |
	[mimeStream atEnd] whileFalse: [
		(nibA := self nextValue) ifNil: [^ dataStream].
		(nibB := self nextValue) ifNil: [^ dataStream].
		dataStream nextPut: ((nibA bitShift: 2) + (nibB bitShift: -4)).
		nibB := nibB bitAnd: 16rF.
		(nibC := self nextValue) ifNil: [^ dataStream].
		dataStream nextPut: ((nibB bitShift: 4) + (nibC bitShift: -2)).
		nibC := nibC bitAnd: 16r3.
		(nibD := self nextValue) ifNil: [^ dataStream].
		dataStream nextPut: ((nibC bitShift: 6) + nibD).
		].
	^ dataStream! !

!Base32MimeConverter methodsFor: 'conversion' stamp: 'UdoSchneider 10/10/2014 23:56'!
mimeEncode
	"Convert from data to 6 bit characters."

	| raw nib lineLength numBytes |
	lineLength := 0.
	numBytes := 0.
	[ dataStream atEnd ]
		whileFalse: [ 
			(multiLine and: [ lineLength >= 70 ])
				ifTrue: [ 
					mimeStream cr.
					lineLength := 0 ].
			data := raw := dataStream next asInteger.
			numBytes := numBytes + 1.
			nib := (data bitAnd: 16rF8) bitShift: -3.
			mimeStream nextPut: (ToCharTable at: nib + 1).
			(raw := dataStream next) ifNil: [ raw := 0 ] ifNotNil: [ numBytes := numBytes + 1 ].
			data := ((data bitAnd: 16r7) bitShift: 8) + raw asInteger.
			nib := (data bitAnd: 16r7C0) bitShift: -6.
			mimeStream nextPut: (ToCharTable at: nib + 1).
			nib := (data bitAnd: 16r3E) bitShift: -1.
			mimeStream nextPut: (ToCharTable at: nib + 1).
			(raw := dataStream next) ifNil: [ raw := 0 ] ifNotNil: [ numBytes := numBytes + 1 ].
			data := ((data bitAnd: 16r1) bitShift: 8) + raw asInteger.
			nib := (data bitAnd: 16r1F0) bitShift: -4.
			mimeStream nextPut: (ToCharTable at: nib + 1).
			(raw := dataStream next) ifNil: [ raw := 0 ] ifNotNil: [ numBytes := numBytes + 1 ].
			data := ((data bitAnd: 16rF) bitShift: 8) + raw asInteger.
			nib := (data bitAnd: 16rF80) bitShift: -7.
			mimeStream nextPut: (ToCharTable at: nib + 1).
			nib := (data bitAnd: 16r7C) bitShift: -2.
			mimeStream nextPut: (ToCharTable at: nib + 1).
			(raw := dataStream next) ifNil: [ raw := 0 ] ifNotNil: [ numBytes := numBytes + 1 ].
			data := ((data bitAnd: 16r3) bitShift: 8) + raw asInteger.
			nib := (data bitAnd: 16r3E0) bitShift: -5.
			mimeStream nextPut: (ToCharTable at: nib + 1).
			nib := data bitAnd: 16r1F.
			mimeStream nextPut: (ToCharTable at: nib + 1).
			lineLength := lineLength + 8 ].
	numBytes \\ 5 > 0
		ifTrue: [ 
			mimeStream skip: (#(-7 -6 -4 -3 -1) at: numBytes \\ 5 + 1).
			(#(7 6 4 3 1) at: numBytes \\ 5 + 1) timesRepeat: [ mimeStream nextPut: $= ] ].
	^ mimeStream! !

!Base32MimeConverter methodsFor: 'accessing' stamp: 'UdoSchneider 10/10/2014 21:18'!
multiLine
	^ multiLine! !

!Base32MimeConverter methodsFor: 'accessing' stamp: 'UdoSchneider 10/10/2014 21:18'!
multiLine: anObject
	multiLine := anObject! !

!Base32MimeConverter methodsFor: 'conversion' stamp: 'UdoSchneider 10/10/2014 21:18'!
nextValue
	"The next six bits of data char from the mimeStream, or nil.  Skip all other chars"
	| raw num |
	[raw := mimeStream next.
	raw ifNil: [^ nil].	"end of stream"
	raw == $= ifTrue: [^ nil].
	num := FromCharTable at: raw asciiValue + 1.
	num ifNotNil: [^ num].
	"else ignore space, return, tab, ..."
	true] whileTrue.! !
Base32MimeConverter initialize!
Base32HexMimeConverter initialize!
